/*
 * Copyright 2001 by Paul Mattes.
 *   Permission to use, copy, modify, and distribute this software and its
 *   documentation for any purpose and without fee is hereby granted,
 *   provided that the above copyright notice appear in all copies and that
 *   both that copyright notice and this permission notice appear in
 *   supporting documentation.
 */

/* 
 * lib3270.c
 *
 *	Glue between lib3270 and the common 3270 functions.
 */

#include "globals.h"

#include <sys/wait.h>
#include <signal.h>
#include <errno.h>
#include <stdarg.h>
#include "appres.h"
#include "3270ds.h"
#include "resources.h"
#include "ctlr.h"

#include "actionsc.h"
#include "ansic.h"
#include "charsetc.h"
#include "ctlrc.h"
#include "ftc.h"
#include "gluec.h"
#include "hostc.h"
#include "keymapc.h"
#include "kybdc.h"
#include "macrosc.h"
#include "menubarc.h"
#include "popupsc.h"
#include "screenc.h"
#include "selectc.h"
#include "tablesc.h"
#include "telnetc.h"
#include "togglesc.h"
#include "trace_dsc.h"
#include "utilc.h"

#include "lib3270.h"

static enum {
	NOT_WAITING,		/* Not waiting */
	AWAITING_CONNECT,	/* Connect (negotiation completion) */
	AWAITING_RESET,		/* Keyboard locked */
	AWAITING_FT,		/* File transfer in progress */
	AWAITING_INPUT,		/* Wait InputField */
	AWAITING_3270,		/* Wait 3270Mode */
	AWAITING_ANSI,		/* Wait NVTMode */
	AWAITING_OUTPUT,	/* Wait Output */
	AWAITING_SOUTPUT,	/* Snap Wait */
	AWAITING_DISCONNECT	/* Wait Disconnect */
} waiting = NOT_WAITING;
static const char *wait_name[] = {
	"not waiting",
	"connection incomplete",
	"keyboard locked",
	"file transfer in progress",
	"need input field",
	"need 3270 mode",
	"need NVT mode",
	"need host output",
	"need snap host output",
	"need host disconnect"
};
static const char *unwait_name[] = {
	"wasn't waiting",
	"connection complete",
	"keyboard unlocked",
	"file transfer complete",
	"input field found",
	"in 3270 mode",
	"in NVT mode",
	"host generated output",
	"host generated snap output",
	"host disconnected"
};
static unsigned long wait_id = 0L;
static Boolean non_blocking = False;
static Boolean init_failed = False;
static int poll_wait(Boolean nonblock);

static int dummy_handle = 0;
static void *dummy_handlep = &dummy_handle;

#define RETVAL_INC	4096	/* allocation unit for return values */

/*
 * Return value structures.
 * There are RETVAL_RING + 1 of them, with the last slot reserved for
 * initialization errors.
 */
static struct {
	const char *action;
	char *buf;
	int len;
	int max;
	char *ptr;
	int cmd_ret;
} retval[RETVAL_RING + 1];
static int rix = RETVAL_RING;

static void
set_return(const char *s)
{
	int sl = strlen(s);
	int inc = retval[rix].max - (retval[rix].len + sl);

	if (inc <= 0) {
		inc = ((-inc + (RETVAL_INC - 1)) / RETVAL_INC) * RETVAL_INC;
		retval[rix].buf =
		    Realloc(retval[rix].buf, retval[rix].max + inc);
		retval[rix].ptr = retval[rix].buf + retval[rix].len;
		retval[rix].max += inc;
	}
	(void) strcpy(retval[rix].ptr, (s));
	retval[rix].ptr += sl;
	retval[rix].len += sl;
}

/* Return the return code and result text from the last command. */
const char *
result_3270(l3h_t handle, int *rcode)
{
	/*
	 * If they supplied a non-NULL handle, and it isn't our canned value,
	 * or there isn't an active connection, then say "bad handle".
	 */
	if (handle != NULL && (handle != dummy_handlep || !dummy_handle)) {
		if (rcode != NULL)
			*rcode = L_BADHANDLE;
		return "invalid handle";
	}

	/* If there is a command in progress, then the caller is "blocked". */
	if (waiting != NOT_WAITING) {
		if (rcode != NULL)
			*rcode = L_BLOCKED;
		return "blocked";
	}

	/*
	 * If the handle is NULL, return any open failure; if not, return
	 * the result of the last command.
	 */
	if (rcode != NULL)
		*rcode = retval[handle? rix: RETVAL_RING].cmd_ret;
	return retval[handle? rix: RETVAL_RING].buf;
}

/* Local prototypes. */
static void ps_clear(void);
static void negotiate(void);
static char *scatv(const char *s);
static void snap_save(void);
static void wait_timed_out(void);

/* Macros.c stuff. */
static Boolean in_cmd = False;
static Boolean output_wait_needed = False;
static char *pending_string = NULL;
static char *pending_string_ptr = NULL;
static Boolean pending_hex = False;
Boolean macro_output = False;

/* Is the keyboard is locked due to user input? */
#define KBWAIT	(kybdlock & (KL_OIA_LOCKED|KL_OIA_TWAIT|KL_DEFERRED_UNLOCK))

/* Is it safe to continue a script waiting for an input field? */
#define INPUT_OKAY ( \
    IN_SSCP || \
    (IN_3270 && formatted && cursor_addr && !KBWAIT) || \
    (IN_ANSI && !(kybdlock & KL_AWAITING_FIRST)) \
)

/* Is is safe to continue a script waiting for the connection to complete? */
#define CONNECT_DONE	(IN_SSCP || IN_3270 || IN_ANSI)

/* Shorthand macro to unlock the current action. */
#define UNBLOCK() { \
	trace_event("Unblocked %s (%s)\n", retval[rix].action, \
	    unwait_name[waiting]); \
	waiting = NOT_WAITING; \
	if (wait_id != 0L) { \
		RemoveTimeOut(wait_id); \
		wait_id = 0L; \
	} \
}

/*
 * Called when the host connects, disconnects, or changes modes.
 * When we connect or change modes, clear the screen.
 * When we disconnect, clear the pending string, so we don't keep trying to
 * feed it to a dead host.
 */
static void
main_connect(Boolean ignored)
{
	if (CONNECTED) {
		ctlr_erase(True);
		/* Check for various wait conditions. */
		switch (waiting) {
		case AWAITING_CONNECT:
			if (CONNECT_DONE)
				UNBLOCK();
			break;
		case AWAITING_3270:
			if (IN_3270)
				UNBLOCK();
			break;
		case AWAITING_ANSI:
			if (IN_ANSI)
				UNBLOCK();
			break;
		default:
			/* Nothing we can figure out here. */
			break;
		}
	} else {
		if (appres.disconnect_clear)
			ctlr_erase(True);
		ps_clear();

		/* Cause (almost) any pending Wait command to fail. */
		if (waiting != NOT_WAITING) {
			if (waiting == AWAITING_DISCONNECT) {
				UNBLOCK();
			} else {
				trace_event("Unblocked %s (was '%s') -- "
					"failure\n", retval[rix].action,
					wait_name[waiting]);
				popup_an_error("Host disconnected");
				waiting = NOT_WAITING;
			}
		}
	}
}

/* Usage. */
void
usage(char *msg)
{
	sms_error(msg);
	init_failed = True;
}

/* Initialization procedure for the library. */
int
open_3270(unsigned long flags, const char **options, l3h_t *h_ret)
{
	int argc;
	const char **argv;
	const char *cl_hostname = CN;

	/* Assume the worst. */
	*h_ret = NULL;

	/* Get ready for output. */
	rix = RETVAL_RING;
	in_cmd = True;
	retval[rix].action = "init";
	retval[rix].ptr = retval[rix].buf;
	retval[rix].len = 0;
	set_return("");

	/* Sorry, only one instance. */
	if (dummy_handle) {
		set_return("too many instances");
		in_cmd = False;
		return (retval[rix].cmd_ret = L_FAIL);
	}

	/* Redirect warnings. */
	Warning_redirect = sms_error;

	/* Process flags. */
	if (flags & NON_BLOCKING)
		non_blocking = True;

	/* Turn options into argc/argv. */
	for (argc = 0; options[argc] != CN; argc++)
		;
	argv = Malloc((argc + 2) * sizeof(char *));
	argv[0] = "lib3270";
	for (argc = 1; options[argc - 1] != CN; argc++) {
		argv[argc] = options[argc - 1];
	}
	argv[argc] = CN;

	argc = parse_command_line(argc, argv, &cl_hostname);
	if (init_failed) {
		in_cmd = False;
		return (retval[rix].cmd_ret = L_FAIL);
	}

	if (charset_init(appres.charset) != CS_OKAY) {
		sms_error(xs_buffer("Cannot find charset \"%s\"",
		    appres.charset));
		(void) charset_init(CN);
	}
	ctlr_init(-1);
	ctlr_reinit(-1);
	kybd_init();
	ansi_init();
#if defined(X3270_FT) /*[*/
	ft_init();
#endif /*]*/

	register_schange(ST_CONNECT, main_connect);
	register_schange(ST_3270_MODE, main_connect);

	/* Make sure we don't fall over any SIGPIPEs. */
	(void) signal(SIGPIPE, SIG_IGN);

	/* Handle initial toggle settings. */
#if defined(X3270_TRACE) /*[*/
	if (!appres.debug_tracing) {
		appres.toggle[DS_TRACE].value = False;
		appres.toggle[EVENT_TRACE].value = False;
	}
#endif /*]*/
	initialize_toggles();

	/* Connect to the host, and wait for negotiation to complete. */
	if (cl_hostname != CN) {
		retval[rix].action = NewString("[initial connection]");
		if (host_connect(cl_hostname) < 0) {
			in_cmd = False;
			return (retval[rix].cmd_ret = L_FAIL);
		}
		if (CONNECTED || HALF_CONNECTED) {
			sms_connect_wait();
			negotiate();
		}
	}

	in_cmd = False;
	dummy_handle = 1;
	if (retval[rix].cmd_ret == L_OK) {
		dummy_handle = 1;
		*h_ret = dummy_handlep;
		return L_OK;
	} else {
		return retval[rix].cmd_ret;
	}
}

int
close_3270(l3h_t handle)
{
	unsigned num_params = 0;

	if (handle != dummy_handlep || !dummy_handle)
		return L_BADHANDLE;

	/* Disconnect, if we need to. */
	(void) Disconnect_action(NULL, NULL, NULL, &num_params);

	/* Get ready for next time. */
	dummy_handle = 0;
	return L_OK;
}


/* Replacements for the logic in macros.c. */


/* Process the pending string (set by the String command). */
static void
process_pending_string(void)
{
	if (pending_string_ptr == NULL || waiting != NOT_WAITING)
		return;

	if (pending_hex) {
		hex_input(pending_string_ptr);
		ps_clear();
	} else {
		int len = strlen(pending_string_ptr);
		int len_left;

		len_left = emulate_input(pending_string_ptr, len, False);
		if (len_left) {
			pending_string_ptr += len - len_left;
			return;
		} else
			ps_clear();
	}
	if (KBWAIT) {
		trace_event("Blocked %s (keyboard locked)\n",
		    retval[rix].action);
		waiting = AWAITING_RESET;
	}
}

/* Clear out the pending string. */
static void
ps_clear(void)
{
	if (pending_string != NULL) {
		pending_string_ptr = NULL;
		Free(pending_string);
		pending_string = NULL;
	}
}

/* The basic wrapper for actions. */
static int
lib3270_cmd(l3h_t handle, XtActionProc proc, const char **params,
    int num_params)
{
	int j;
	unsigned count = num_params;

	/* Gotta have a valid handle. */
	if (handle != dummy_handlep || !dummy_handle)
		return L_BADHANDLE;

	/* Can't nest. */
	if (in_cmd)
		return L_NESTED;

	/* Set up ugly global variables. */
	in_cmd = True;
	rix = (rix + 1) % RETVAL_RING;
	retval[rix].action = action_name(proc);
	retval[rix].ptr = retval[rix].buf;
	retval[rix].len = 0;
	set_return("");

	/* Synchronously run any pending I/O's and timeouts.  Ugly. */
	while (process_events(False))
		;

#if defined(X3270_TRACE) /*[*/
	/* Trace what we're about to do. */
	if (toggled(EVENT_TRACE)) {
		trace_event("Running %s", retval[rix].action);
		for (j = 0; j < count; j++) {
			char *s = scatv(params[j]);

			trace_event(" %s", s);
			Free(s);
		}
		trace_event("\n");
	}
#endif /*]*/

	/* Set up more ugly global variables and run the action. */
	ia_cause = IA_SCRIPT;
	retval[rix].cmd_ret = L_OK;
	(*proc)((Widget)NULL, (XEvent *)NULL, (char **)params, &count);

	/* Set implicit wait state. */
#if defined(X3270_FT) /*[*/
	if (ft_state != FT_NONE)
		waiting = AWAITING_FT;
	else
#endif /*]*/
	if (KBWAIT)
		waiting = AWAITING_RESET;

	if (waiting != NOT_WAITING) {
		trace_event("Blocked %s (%s)\n", retval[rix].action,
		    wait_name[waiting]);
		if (non_blocking)
			return L_BLOCKED;
	}

	return poll_wait(False);
}

/* Check for command completion. */
int
poll_3270(l3h_t handle)
{
	if (handle != dummy_handlep || !dummy_handle)
		return L_BADHANDLE;
	return poll_wait(True);
}

/* Wait for command completion. */
int
wait_3270(l3h_t handle)
{
	if (handle != dummy_handlep || !dummy_handle)
		return L_BADHANDLE;
	return poll_wait(False);
}

/* Check completion status of a blocked command. */
static int
poll_wait(Boolean nonblock)
{
	/* Make sure we're actually waiting for something. */
	if (!in_cmd)
		return L_OK;

	/*
	 * Process responses and push any pending string, until
	 * we can proceed.
	 */
	process_pending_string();
	while (waiting != NOT_WAITING) {

		/* Process pending file I/O. */
		if (nonblock) {
			while (process_events(False))
				;
		} else {
			(void) process_events(True);
		}

		/*
		 * Check for the completion of output-related wait conditions.
		 */
		switch (waiting) {
		case AWAITING_INPUT:
			if (INPUT_OKAY)
				UNBLOCK();
			break;
		case AWAITING_RESET:
			if (!KBWAIT)
				UNBLOCK();
			break;
#if defined(X3270_FT) /*[*/
		case AWAITING_FT:
			if (ft_state == FT_NONE)
				UNBLOCK();
			break;
#endif /*]*/
		default:
			break;
		}

		/* Push more string text in. */
		process_pending_string();

		/* If we're not allowed to block, we're done. */
		if (nonblock && waiting != NOT_WAITING)
			return L_BLOCKED;
	}
#if defined(X3270_TRACE) /*[*/
	if (toggled(EVENT_TRACE)) {
#		define TRUNC_LEN 40
		char s_trunc[TRUNC_LEN + 1];

		trace_event("Completed %s (%s)", retval[rix].action,
			(retval[rix].cmd_ret == L_OK) ? "ok" : "error");
		if (retval[rix].buf[0]) {
			strncpy(s_trunc, retval[rix].buf, TRUNC_LEN);
			s_trunc[TRUNC_LEN] = '\0';
			trace_event(" -> \"");
			fcatv(tracef, s_trunc);
			trace_event("\"");
			if (strlen(retval[rix].buf) > TRUNC_LEN)
				trace_event("...(%d chars)", strlen(retval[rix].buf));
		}
		trace_event("\n");
	}
#endif /*]*/
	in_cmd = False;
	return retval[rix].cmd_ret;
}

static int
lib3270_cmdl(l3h_t handle, XtActionProc proc, ...)
{
	const char *params[64];
	int num_params = 0;
	va_list args;

	va_start(args, proc);
	for (;;) {
		params[num_params] = va_arg(args, const char *);
		if (params[num_params] == CN)
			break;
		num_params++;
	}
	va_end(args);
	params[num_params] = CN;
	return lib3270_cmd(handle, proc, params, num_params);
}

/* select() set-up. */
int
selectvals_3270(l3h_t handle, int *nfds, fd_set *readfds,
    fd_set *writefds, fd_set *exceptfds, struct timeval **timeout,
    struct timeval *timebuf)
{
	if (handle != dummy_handlep || !dummy_handle)
		return L_BADHANDLE;

	return select_setup(nfds, readfds, writefds, exceptfds, timeout,
	    timebuf);
}

/* Do initial connect negotiation. */
static void
negotiate(void)
{
	while (KBWAIT || (waiting == AWAITING_CONNECT && !CONNECT_DONE)) {
		(void) process_events(True);
		if (!PCONNECTED)
			exit(1);
	}
}

/*
 * Indicates whether errors should go to stderr, or be returned to the library.
 */
Boolean
sms_redirect(void)
{
	return in_cmd;
}

/* Returns an error to the library. */
void
sms_error(const char *s)
{
	set_return(s);
	retval[rix].cmd_ret = L_FAIL;
}

/* For now, a no-op.  Used to implement 'Expect'. */
void
sms_store(unsigned char c)
{
}

/* Set the pending string.  Used by the 'String' action. */
void
ps_set(char *s, Boolean is_hex)
{
	pending_string = pending_string_ptr = s;
	pending_hex = is_hex;
}

/* Signal a new connection. */
void
sms_connect_wait(void)
{
	waiting = AWAITING_CONNECT;
}

/* Signal host output. */
void
sms_host_output(void)
{
	/* Release the script, if it is waiting now. */
	switch (waiting) {
	    case AWAITING_SOUTPUT:
		snap_save();
		/* fall through... */
	    case AWAITING_OUTPUT:
		UNBLOCK();
		break;
	    default:
		break;
	}

	/* If there was no script waiting, ensure that it won't later. */
	output_wait_needed = False;
}

/* More no-ops. */
void
login_macro(char *s)
{
}
void
sms_continue(void)
{
}

/* Data query actions. */

static void
dump_range(int first, int len, Boolean in_ascii, unsigned char *buf,
    int rel_rows unused, int rel_cols)
{
	register int i;

	/*
	 * The client has now 'looked' at the screen, so should they later
	 * execute 'Wait(output)', they will actually need to wait for output
	 * from the host.  output_wait_needed is cleared by sms_host_output,
	 * which is called from the write logic in ctlr.c.
	 */
	if (buf == screen_buf)
		output_wait_needed = True;

	for (i = 0; i < len; i++) {
		/* Check for a new row. */
		if (i && !((first + i) % rel_cols)) {
			/* Done with this row. */
			set_return("\n");
		}
		if (in_ascii) {
			char c[2];

			c[0] = cg2asc[buf[first + i]];
			if (!c[0])
				c[0] = ' ';
			c[1] = '\0';
			set_return(c);
		} else {
			char s[5];

			(void) sprintf(s, "0x%02x",
				cg2ebc[buf[first + i]]);
			set_return(s);
		}
	}
}

static void
dump_rectangle(int start_row, int start_col, int rows, int cols,
    Boolean in_ascii, unsigned char *buf, int rel_cols)
{
	register int r, c;

	/*
	 * The client has now 'looked' at the screen, so should they later
	 * execute 'Wait(output)', they will actually need to wait for output
	 * from the host.  output_wait_needed is cleared by sms_host_output,
	 * which is called from the write logic in ctlr.c.
	 */
	if (buf == screen_buf)
		output_wait_needed = True;

	if (!rows || !cols)
		return;

	for (r = start_row; r < start_row + rows; r++) {
		/* New row. */
		set_return("\n");

		for (c = start_col; c < start_col + cols; c++) {
			int loc = (r * rel_cols) + c;

			if (in_ascii) {
				unsigned char ch[2];

				ch[0] = cg2asc[buf[loc]];
				if (!ch[0])
					ch[0] = ' ';
				ch[1] = '\0';
				set_return((char *)ch);
			} else {
				char s[5];

				(void) sprintf(s, "0x%02x",
					cg2ebc[buf[loc]]);
				set_return(s);
			}

		}
	}
}

static void
dump_fixed(String params[], Cardinal count, const char *name, Boolean in_ascii,
	unsigned char *buf, int rel_rows, int rel_cols, int caddr)
{
	int row, col, len, rows = 0, cols = 0;

	switch (count) {
	    case 0:	/* everything */
		row = 0;
		col = 0;
		len = rel_rows*rel_cols;
		break;
	    case 1:	/* from cursor, for n */
		row = caddr / rel_cols;
		col = caddr % rel_cols;
		len = atoi(params[0]);
		break;
	    case 3:	/* from (row,col), for n */
		row = atoi(params[0]);
		col = atoi(params[1]);
		len = atoi(params[2]);
		break;
	    case 4:	/* from (row,col), for rows x cols */
		row = atoi(params[0]);
		col = atoi(params[1]);
		rows = atoi(params[2]);
		cols = atoi(params[3]);
		len = 0;
		break;
	    default:
		popup_an_error("%s requires 0, 1, 3 or 4 arguments", name);
		return;
	}

	if (
	    (row < 0 || row > rel_rows || col < 0 || col > rel_cols || len < 0) ||
	    ((count < 4)  && ((row * rel_cols) + col + len > rel_rows * rel_cols)) ||
	    ((count == 4) && (cols < 0 || rows < 0 ||
			      col + cols > rel_cols || row + rows > rel_rows))
	   ) {
		popup_an_error("%s: Invalid argument", name);
		return;
	}
	if (count < 4)
		dump_range((row * rel_cols) + col, len, in_ascii, buf,
			rel_rows, rel_cols);
	else
		dump_rectangle(row, col, rows, cols, in_ascii, buf, rel_cols);
}

static void
dump_field(Cardinal count, const char *name, Boolean in_ascii)
{
	unsigned char *fa;
	int start, baddr;
	int len = 0;

	if (count != 0) {
		popup_an_error("%s requires 0 arguments", name);
		return;
	}
	if (!formatted) {
		popup_an_error("%s: Screen is not formatted", name);
		return;
	}
	fa = get_field_attribute(cursor_addr);
	start = fa - screen_buf;
	INC_BA(start);
	baddr = start;
	do {
		if (IS_FA(screen_buf[baddr]))
			break;
		len++;
		INC_BA(baddr);
	} while (baddr != start);
	dump_range(start, len, in_ascii, screen_buf, ROWS, COLS);
}

void
Ascii_action(Widget w unused, XEvent *event unused, String *params,
    Cardinal *num_params)
{
	dump_fixed(params, *num_params, action_name(Ascii_action), True,
		screen_buf, ROWS, COLS, cursor_addr);
}

void
AsciiField_action(Widget w unused, XEvent *event unused, String *params unused,
    Cardinal *num_params)
{
	dump_field(*num_params, action_name(AsciiField_action), True);
}

void
Ebcdic_action(Widget w unused, XEvent *event unused, String *params,
    Cardinal *num_params)
{
	dump_fixed(params, *num_params, action_name(Ebcdic_action), False,
		screen_buf, ROWS, COLS, cursor_addr);
}

void
EbcdicField_action(Widget w unused, XEvent *event unused,
    String *params unused, Cardinal *num_params)
{
	dump_field(*num_params, action_name(EbcdicField_action), False);
}

/* "Status" action, returns the s3270 prompt. */
static char *
status_string(void)
{
	char kb_stat;
	char fmt_stat;
	char prot_stat;
	char *connect_stat = NULL;
	char em_mode;
	char s[1024];
	char *r;

	if (!kybdlock)
		kb_stat = 'U';
	else if (!CONNECTED || KBWAIT)
		kb_stat = 'L';
	else
		kb_stat = 'E';

	if (formatted)
		fmt_stat = 'F';
	else
		fmt_stat = 'U';

	if (!formatted)
		prot_stat = 'U';
	else {
		unsigned char *fa;

		fa = get_field_attribute(cursor_addr);
		if (FA_IS_PROTECTED(*fa))
			prot_stat = 'P';
		else
			prot_stat = 'U';
	}

	if (CONNECTED)
		connect_stat = xs_buffer("C(%s)", current_host);
	else
		connect_stat = NewString("N");

	if (CONNECTED) {
		if (IN_ANSI) {
			extern int linemode; /* XXX */
			if (linemode)
				em_mode = 'L';
			else
				em_mode = 'C';
		} else if (IN_SSCP)
			em_mode = 'S';
		else if (IN_3270)
			em_mode = 'I';
		else
			em_mode = 'P';
	} else
		em_mode = 'N';

	(void) sprintf(s, "%c %c %c %s %c %d %d %d %d %d",
	    kb_stat,
	    fmt_stat,
	    prot_stat,
	    connect_stat,
	    em_mode,
	    model_num,
	    ROWS, COLS,
	    cursor_addr / COLS, cursor_addr % COLS);
	r = NewString(s);
	Free(connect_stat);
	return r;
}

void
Status_action(Widget w unused, XEvent *event unused, String *params,
    Cardinal *num_params)
{
	char *s;

	s = status_string();
	set_return(s);
	Free(s);
}

/*
 * "Snap" action, maintains a snapshot for consistent multi-field comparisons:
 *
 *  Snap Save
 *	updates the saved image from the live image
 *  Snap Rows
 *	returns the number of rows
 *  Snap Cols
 *	returns the number of columns
 *  Snap Staus
 *  Snap Ascii ...
 *  Snap AsciiField (not yet)
 *  Snap Ebcdic ...
 *  Snap EbcdicField (not yet)
 *	runs the named command
 *  Snap Wait [tmo] Output
 *	waits for the screen to change
 */

static char *snap_status = NULL;
static unsigned char *snap_buf = NULL;
static int snap_rows = 0;
static int snap_cols = 0;
static int snap_field_start = -1;
static int snap_field_length = -1;
static int snap_caddr = 0;

static void
snap_save(void)
{
	output_wait_needed = True;
	if (snap_status != NULL)
		Free(snap_status);
	snap_status = status_string();

	if (snap_buf != NULL)
		Free(snap_buf);
	snap_buf = (unsigned char *)Malloc(ROWS*COLS);
	(void) memcpy(snap_buf, screen_buf, ROWS*COLS);

	snap_rows = ROWS;
	snap_cols = COLS;

	if (!formatted) {
		snap_field_start = -1;
		snap_field_length = -1;
	} else {
		unsigned char *fa;
		int baddr;

		snap_field_length = 0;
		fa = get_field_attribute(cursor_addr);
		snap_field_start = fa - screen_buf;
		INC_BA(snap_field_start);
		baddr = snap_field_start;
		do {
			if (IS_FA(screen_buf[baddr]))
				break;
			snap_field_length++;
			INC_BA(baddr);
		} while (baddr != snap_field_start);
	}
	snap_caddr = cursor_addr;
}

void
Snap_action(Widget w unused, XEvent *event unused, String *params,
    Cardinal *num_params)
{
	char nbuf[16];

	if (*num_params == 0) {
		snap_save();
		return;
	}

	/* Handle 'Snap Wait' separately. */
	if (!strcasecmp(params[0], action_name(Wait_action))) {
		unsigned long tmo;
		char *ptr;
		int maxp = 0;

		if (*num_params > 1 &&
		    (tmo = strtoul(params[1], &ptr, 10)) > 0 &&
		    ptr != params[0] &&
		    *ptr == '\0') {
			maxp = 3;
		} else {
			tmo = 0L;
			maxp = 2;
		}
		if (*num_params > maxp) {
			popup_an_error("Too many arguments to %s %s",
			    action_name(Snap_action),
			    action_name(Wait_action));
			    return;
		}
		if (*num_params < maxp) {
			popup_an_error("Too few arguments to %s %s",
			    action_name(Snap_action),
			    action_name(Wait_action));
			    return;
		}
		if (strcasecmp(params[*num_params - 1], "Output")) {
			popup_an_error("Unknown parameter to %s %s",
			action_name(Snap_action),
			action_name(Wait_action));
			return;
		}

		/* Must be connected. */
		if (!(CONNECTED || HALF_CONNECTED)) {
			popup_an_error("%s: Not connected",
			    action_name(Snap_action));
			return;
		}

		/*
		 * Make sure we need to wait.
		 * If we don't, then Snap Wait Output is equivalen to Snap Save.
		 */
		if (!output_wait_needed) {
			snap_save();
			return;
		}

		/* Set the new state. */
		waiting = AWAITING_SOUTPUT;

		/* Set up a timeout, if they want one. */
		if (tmo)
			wait_id = AddTimeOut(tmo * 1000, wait_timed_out);
		return;
	}

	if (!strcasecmp(params[0], "Save")) {
		if (*num_params != 1) {
			popup_an_error("Extra argument(s)");
			return;
		}
		snap_save();
	} else if (!strcasecmp(params[0], "Status")) {
		if (*num_params != 1) {
			popup_an_error("Extra argument(s)");
			return;
		}
		if (snap_status == NULL) {
			popup_an_error("No saved state");
			return;
		}
		set_return(snap_status);
	} else if (!strcasecmp(params[0], "Rows")) {
		if (*num_params != 1) {
			popup_an_error("Extra argument(s)");
			return;
		}
		if (snap_status == NULL) {
			popup_an_error("No saved state");
			return;
		}
		(void) sprintf(nbuf, "%d", snap_rows);
		set_return(nbuf);
	} else if (!strcasecmp(params[0], "Cols")) {
		if (*num_params != 1)
			popup_an_error("extra argument(s)");
		(void) sprintf(nbuf, "%d", snap_cols);
		set_return(nbuf);
	} else if (!strcasecmp(params[0], action_name(Ascii_action))) {
		if (snap_status == NULL) {
			popup_an_error("No saved state");
			return;
		}
		dump_fixed(params + 1, *num_params - 1,
			action_name(Ascii_action), True, snap_buf,
			snap_rows, snap_cols, snap_caddr);
	} else if (!strcasecmp(params[0], action_name(Ebcdic_action))) {
		if (snap_status == NULL) {
			popup_an_error("No saved state");
			return;
		}
		dump_fixed(params + 1, *num_params - 1,
			action_name(Ebcdic_action), False, snap_buf,
			snap_rows, snap_cols, snap_caddr);
	} else {
		popup_an_error("%s: Argument '%s' must be Save, Status, Rows, Cols, "
		    "%s, %s or %s",
		    action_name(Snap_action),
		    params[0],
		    action_name(Wait_action),
		    action_name(Ascii_action),
		    action_name(Ebcdic_action));
	}
}

static void
wait_timed_out(void)
{
	popup_an_error("Wait timed out");
	wait_id = 0L;
	UNBLOCK();
}

void
Wait_action(Widget w unused, XEvent *event unused, String *params,
    Cardinal *num_params)
{
	unsigned long tmo = 0;
	char *ptr;
	Cardinal np;
	String *pr;

	if (*num_params > 0 &&
	    (tmo = strtoul(params[0], &ptr, 10)) > 0 &&
	     ptr != params[0] &&
	     *ptr == '\0') {
		np = *num_params - 1;
		pr = params + 1;
	 } else {
		np = *num_params;
		pr = params;
	}

	if (np == 0) {
		if (!CONNECTED) {
			popup_an_error("Not connected");
			return;
		}
		if (!INPUT_OKAY)
			waiting = AWAITING_INPUT;
		return;
	}
	if (np != 1) {
		popup_an_error("Too many parameters");
		return;
	}
	if (!strcasecmp(pr[0], "InputField")) {
		/* Same as no parameters. */
		if (!CONNECTED) {
			popup_an_error("Not connected");
			return;
		}
		if (!INPUT_OKAY)
			waiting = AWAITING_INPUT;
	} else if (!strcasecmp(pr[0], "Output")) {
		if (!CONNECTED) {
			popup_an_error("Not connected");
			return;
		}
		if (output_wait_needed)
			waiting = AWAITING_OUTPUT;
	} else if (!strcasecmp(pr[0], "3270") ||
		   !strcasecmp(pr[0], "3270Mode")) {
		if (!CONNECTED) {
			popup_an_error("Not connected");
			return;
		}
		if (!IN_3270)
			waiting = AWAITING_3270;
	} else if (!strcasecmp(pr[0], "ansi") ||
		   !strcasecmp(pr[0], "NVTMode")) {
		if (!CONNECTED) {
			popup_an_error("Not connected");
			return;
		}
		if (!IN_ANSI)
			waiting = AWAITING_ANSI;
	} else if (!strcasecmp(pr[0], "Disconnect")) {
		if (CONNECTED)
			waiting = AWAITING_DISCONNECT;
	} else {
		popup_an_error("Unknown Wait type: %s", pr[0]);
		return;
	}

	if (tmo)
		wait_id = AddTimeOut(tmo * 1000L, wait_timed_out);
}

/* Generate a response to a script command. */
void
sms_info(const char *fmt, ...)
{
	va_list args;
	char buf[4096];

	va_start(args, fmt);
	(void) vsprintf(buf, fmt, args);
	va_end(args);
	set_return(buf);
}

/* Like fcatv, but goes to a buffer. */
static char *
scatv(const char *s)
{
#define ALLOC_INC	1024
	char *buf;
	int buflen;
	int bufused = 0;
	char c;
#define add_space(n) \
		if (bufused + (n) >= buflen) { \
			buflen += ALLOC_INC; \
			buf = Realloc(buf, buflen); \
		} \
		bufused += (n);

	buf = Malloc(ALLOC_INC);
	buflen = ALLOC_INC;
	*buf = '\0';

	while ((c = *s++))  {
		switch (c) {
		case '\n':
			add_space(2);
			(void) strcat(buf, "\\n");
			break;
		case '\t':
			add_space(2);
			(void) strcat(buf, "\\t");
			break;
		case '\b':
			add_space(2);
			(void) strcat(buf, "\\b");
			break;
		case '\f':
			add_space(2);
			(void) strcat(buf, "\\f");
			break;
		case ' ':
			add_space(2);
			(void) strcat(buf, "\\ ");
			break;
		default:
			if ((c & 0x7f) < ' ') {
				add_space(4);
				(void) sprintf(buf + bufused, "\\%03o",
					c & 0xff);
				break;
			} else {
				add_space(1);
				*(buf + bufused - 1) = c;
				*(buf + bufused) = '\0';
			}
		}
	}
	return buf;
#undef add_space
}

/* Individual actions. */

static const char *
nstr(int n)
{
	static char rbuf[4][64];
	static int rptr = 0;
	char *ret;

	(void) sprintf(ret = rbuf[rptr], "%d", n);
	rptr = (rptr + 1) % 4;
	return ret;
}

#define TRIVIAL_INT_ACTION(mode, n) \
    int \
    mode ## n(l3h_t handle) \
    { \
	    return lib3270_cmdl(handle, n ## _action, CN); \
    }

#define ERRCN(x)	(((x) < 0)? CN: retval[rix].buf)

#define TRIVIAL_STRING_ACTION(mode, n) \
    const char * \
    mode ## n(l3h_t handle) \
    { \
	    return ERRCN(lib3270_cmdl(handle, n ## _action, CN)); \
    }

#define TRIVIAL_CHARPARM_ACTION(mode, n) \
    int \
    mode ## n(l3h_t handle, const char *s) \
    { \
	    return lib3270_cmdl(handle, n ## _action, s, CN); \
    }

const char *
sAsciiAll(l3h_t handle)
{
	return ERRCN(lib3270_cmdl(handle, Ascii_action, CN));
}

const char *
sAsciiFromCursor(l3h_t handle, int len)
{
	return ERRCN(lib3270_cmdl(handle, Ascii_action, nstr(len), CN));
}

const char *
sAsciiLine(l3h_t handle, int row, int col, int len)
{
	return ERRCN(lib3270_cmdl(handle, Ascii_action, nstr(row), nstr(col),
	    nstr(len), CN));
}

const char *
sAsciiBlock(l3h_t handle, int row, int col, int rows, int cols)
{
	return ERRCN(lib3270_cmdl(handle, Ascii_action, nstr(row), nstr(col),
	    nstr(rows), nstr(cols), CN));
}

TRIVIAL_STRING_ACTION(s, AsciiField)
TRIVIAL_INT_ACTION(s, Attn)
TRIVIAL_INT_ACTION(s, BackSpace)
TRIVIAL_INT_ACTION(s, BackTab)
TRIVIAL_INT_ACTION(a, Clear)

int
sCols(l3h_t handle)
{
	if (handle != dummy_handlep || !dummy_handle)
		return L_BADHANDLE;
	return COLS;
}

TRIVIAL_CHARPARM_ACTION(a, Connect)
TRIVIAL_INT_ACTION(a, CursorSelect)
TRIVIAL_INT_ACTION(s, Delete)
TRIVIAL_INT_ACTION(s, DeleteField)
TRIVIAL_INT_ACTION(s, DeleteWord)
TRIVIAL_INT_ACTION(s, Disconnect)
TRIVIAL_INT_ACTION(s, Down)
TRIVIAL_INT_ACTION(s, Dup)

const char *
aEbcdicAll(l3h_t handle)
{
	return ERRCN(lib3270_cmdl(handle, Ebcdic_action, CN));
}

const char *
aEbcdicFromCursor(l3h_t handle, int len)
{
	return ERRCN(lib3270_cmdl(handle, Ebcdic_action, nstr(len), CN));
}

const char *
aEbcdicLine(l3h_t handle, int row, int col, int len)
{
	return ERRCN(lib3270_cmdl(handle, Ebcdic_action, nstr(row), nstr(col),
	    nstr(len), CN));
}

const char *
aEbcdicBlock(l3h_t handle, int row, int col, int rows, int cols)
{
	char nbuf1[64], nbuf2[64], nbuf3[64], nbuf4[64];

	(void) sprintf(nbuf1, "%d", row);
	(void) sprintf(nbuf2, "%d", col);
	(void) sprintf(nbuf3, "%d", rows);
	(void) sprintf(nbuf4, "%d", cols);
	return ERRCN(lib3270_cmdl(handle, Ebcdic_action, nstr(row), nstr(col),
	    nstr(rows), nstr(cols), CN));
}

TRIVIAL_STRING_ACTION(s, EbcdicField)
TRIVIAL_INT_ACTION(a, Enter)
TRIVIAL_INT_ACTION(s, Erase)
TRIVIAL_INT_ACTION(s, EraseEOF)
TRIVIAL_INT_ACTION(s, EraseInput)
TRIVIAL_INT_ACTION(s, FieldEnd)
TRIVIAL_INT_ACTION(s, FieldMark)
TRIVIAL_INT_ACTION(s, FieldExit)
TRIVIAL_CHARPARM_ACTION(a, HexString)
TRIVIAL_INT_ACTION(s, Home)
TRIVIAL_INT_ACTION(s, Insert)
TRIVIAL_INT_ACTION(s, Interrupt)
TRIVIAL_CHARPARM_ACTION(a, Key)
TRIVIAL_INT_ACTION(s, Left)
TRIVIAL_INT_ACTION(s, Left2)
TRIVIAL_INT_ACTION(s, MonoCase)

int
sMoveCursor(l3h_t handle, int row, int col)
{
	return lib3270_cmdl(handle, MoveCursor_action, nstr(row), nstr(col),
	    CN);
}

TRIVIAL_INT_ACTION(s, Newline)
TRIVIAL_INT_ACTION(s, NextWord)

int
aPA(l3h_t handle, int n)
{
	return lib3270_cmdl(handle, PA_action, nstr(n), CN);
}

int
aPF(l3h_t handle, int n)
{
	return lib3270_cmdl(handle, PF_action, nstr(n), CN);
}

TRIVIAL_INT_ACTION(s, PreviousWord)
TRIVIAL_INT_ACTION(s, Reset)
TRIVIAL_INT_ACTION(s, Right)
TRIVIAL_INT_ACTION(s, Right2)

int
sRows(l3h_t handle)
{
	if (handle != dummy_handlep || !dummy_handle)
		return L_BADHANDLE;
	return ROWS;
}

int
sSnapSave(l3h_t handle)
{
	return lib3270_cmdl(handle, Snap_action, CN);
}

const char *
sSnapStatus(l3h_t handle)
{
	return ERRCN(lib3270_cmdl(handle, Snap_action, "Status", CN));
}

int
sSnapRows(l3h_t handle)
{
	return (lib3270_cmdl(handle, Snap_action, "Rows", CN) < 0)? -1:
	    atoi(retval[rix].buf);
}

int
sSnapCols(l3h_t handle)
{
	return (lib3270_cmdl(handle, Snap_action, "Cols", CN) < 0)? -1:
	    atoi(retval[rix].buf);
}

int
aSnapWaitOutput(l3h_t handle, int tmo)
{
	if (tmo > 0)
		return lib3270_cmdl(handle, Snap_action, "Wait", nstr(tmo),
		    "Output", CN);
	else
		return lib3270_cmdl(handle, Snap_action, "Wait", "Output", CN);
}

const char *
sSnapAsciiAll(l3h_t handle)
{
	return ERRCN(lib3270_cmdl(handle, Snap_action, "Ascii", CN));
}

const char *
sSnapAsciiFromCursor(l3h_t handle, int len)
{
	return ERRCN(lib3270_cmdl(handle, Snap_action, "Ascii", nstr(len), CN));
}

const char *
sSnapAsciiLine(l3h_t handle, int row, int col, int len)
{
	return ERRCN(lib3270_cmdl(handle, Snap_action, "Ascii", nstr(row),
	    nstr(col), nstr(len), CN));
}

const char *
sSnapAsciiBlock(l3h_t handle, int row, int col, int rows, int cols)
{
	return ERRCN(lib3270_cmdl(handle, Snap_action, "Ascii", nstr(row),
	    nstr(col), nstr(rows), nstr(cols), CN));
}

const char *
sSnapEbcdicAll(l3h_t handle)
{
	return ERRCN(lib3270_cmdl(handle, Snap_action, "Ebcdic", CN));
}

const char *
sSnapEbcdicFromCursor(l3h_t handle, int len)
{
	return ERRCN(lib3270_cmdl(handle, Snap_action, "Ebcdic", nstr(len),
	    CN));
}

const char *
sSnapEbcdicLine(l3h_t handle, int row, int col, int len)
{
	return ERRCN(lib3270_cmdl(handle, Snap_action, "Ebcdic", nstr(row),
	    nstr(col), nstr(len), CN));
}

const char *
sSnapEbcdicBlock(l3h_t handle, int row, int col, int rows, int cols)
{
	return ERRCN(lib3270_cmdl(handle, Snap_action, "Ebcdic", nstr(row),
	    nstr(col), nstr(rows), nstr(cols), CN));
}

const char *
sStatus(l3h_t handle)
{
	return ERRCN(lib3270_cmdl(handle, Status_action, CN));
}

TRIVIAL_CHARPARM_ACTION(a, String)
TRIVIAL_INT_ACTION(a, SysReq)
TRIVIAL_INT_ACTION(s, Tab)
TRIVIAL_INT_ACTION(s, ToggleInsert)
TRIVIAL_INT_ACTION(s, ToggleReverse)

int
aTransfer(l3h_t handle, int argc, const char *argv[])
{
	return lib3270_cmd(handle, Transfer_action, argv, argc);
}

TRIVIAL_INT_ACTION(s, Up)

#define WAIT_ACTION(ty) \
    int \
    aWait ## ty(l3h_t handle, int tmo) \
    { \
	    if (tmo > 0) { \
		    return lib3270_cmdl(handle, Wait_action, #ty, nstr(tmo), CN); \
	    } else { \
		    return lib3270_cmdl(handle, Wait_action, #ty, CN); \
	    } \
    }

WAIT_ACTION(InputField)
WAIT_ACTION(Output)
WAIT_ACTION(3270Mode)
WAIT_ACTION(NVTMode)
WAIT_ACTION(Disconnect)
