/*
 * Copyright 2001 by Paul Mattes.
 *   Permission to use, copy, modify, and distribute this software and its
 *   documentation for any purpose and without fee is hereby granted,
 *   provided that the above copyright notice appear in all copies and that
 *   both that copyright notice and this permission notice appear in
 *   supporting documentation.
 */

/*
 * lib3270.h
 *
 *	Header file for the 3270 C library.
 */

#if !defined(__LIB_3270_H__) /*[*/
#define __LIB_3270_H__

/*
 * lib3270 instance handle.  To be used when (eventually) lib3270 supports
 * multiple connections.
 */
typedef void *l3h_t;

/*
 * Return status.
 */
#define L_OK		0		/* success */
#define L_FAIL		(-1)		/* failure */
#define	L_BLOCKED	(-2)		/* incomplete (non-blocking mode) */
#define L_NESTED	(-3)		/* illegal nesting */
#define L_BADHANDLE	(-4)		/* invalid handle */

/*
 * Initialization function.
 */
extern int open_3270(unsigned long flags, const char **options,
    l3h_t *handle_ret);
#define NON_BLOCKING	0x1		/* non-blocking mode */

/*
 * Termination function.
 */
extern int close_3270(l3h_t handle);

/*
 * Completion polling functions.
 */
extern int poll_3270(l3h_t handle);
extern int wait_3270(l3h_t handle);

/* Number of buffers saved before they are overwritten. */
#define RETVAL_RING	10

/* Result from last operation (error message if it failed). */
extern const char *result_3270(l3h_t handle, int *rcode);

/*
 * Non-blocking setup function.
 * Returns nfds and file descriptors; modifies timeval for the event(s) that
 * the library is interested in.  Returns 1 if there are events pending, 0 if
 * not.
 */
extern int
selectvals_3270(l3h_t handle, int *nfds, fd_set *readfds, fd_set *writefds,
    fd_set *exceptfds, struct timeval **timeout, struct timeval *timebuf);

/*
 * Actions.
 * (const char *) actions return non-NULL for success, NULL for failure.
 *  None of them can block.
 * (int) actions return L_XXX status, above.
 * Returned buffers are from a pool, which is overwritten every RETVAL_RING
 *  calls.
 */

extern const char *sAsciiAll(l3h_t handle);
extern const char *sAsciiFromCursor(l3h_t handle, int len);
extern const char *sAsciiLine(l3h_t handle, int row, int col, int len);
extern const char *sAsciiBlock(l3h_t handle, int row, int col, int rows,
    int cols);
extern const char *sAsciiField(l3h_t handle);
extern int         sAttn(l3h_t handle);
extern int         sBackSpace(l3h_t handle);
extern int         sBackTab(l3h_t handle);
extern int         aClear(l3h_t handle);
extern int         sCols(l3h_t handle);
extern int         aConnect(l3h_t handle, const char *);
extern int         aCursorSelect(l3h_t handle);
extern int         sDelete(l3h_t handle);
extern int         sDeleteField(l3h_t handle);
extern int         sDeleteWord(l3h_t handle);
extern int         sDisconnect(l3h_t handle);
extern int         sDown(l3h_t handle);
extern int         sDup(l3h_t handle);
extern const char *sEbcdicAll(l3h_t handle);
extern const char *sEbcdicFromCursor(l3h_t handle, int len);
extern const char *sEbcdicLine(l3h_t handle, int row, int col, int len);
extern const char *sEbcdicBlock(l3h_t handle, int row, int col, int rows,
    int cols);
extern const char *sEbcdicField(l3h_t handle);
extern int         aEnter(l3h_t handle);
extern int         sErase(l3h_t handle);
extern int         sEraseEOF(l3h_t handle);
extern int         sEraseInput(l3h_t handle);
extern int         sFieldEnd(l3h_t handle);
extern int         sFieldMark(l3h_t handle);
extern int         sFieldExit(l3h_t handle);
extern int         aHexString(l3h_t handle, const char *);
extern int         sHome(l3h_t handle);
extern int         sInsert(l3h_t handle);
extern int         sInterrupt(l3h_t handle);
extern int         aKey(l3h_t handle, const char *);
extern int         sLeft(l3h_t handle);
extern int         sLeft2(l3h_t handle);
extern int         sMonoCase(l3h_t handle);
extern int         sMoveCursor(l3h_t handle, int row, int col);
extern int         sNewline(l3h_t handle);
extern int         sNextWord(l3h_t handle);
extern int         aPA(l3h_t handle, int n);
extern int         aPF(l3h_t handle, int n);
extern int         sPreviousWord(l3h_t handle);
extern int         sReset(l3h_t handle);
extern int         sRight(l3h_t handle);
extern int         sRight2(l3h_t handle);
extern int         sRows(l3h_t handle);
extern int         sSnapSave(l3h_t handle);
extern const char *sSnapStatus(l3h_t handle);
extern int         sSnapRows(l3h_t handle);
extern int         sSnapCols(l3h_t handle);
extern int         aSnapWaitOutput(l3h_t handle, int tmo);
extern const char *sSnapAsciiAll(l3h_t handle);
extern const char *sSnapAsciiFromCursor(l3h_t handle, int len);
extern const char *sSnapAsciiLine(l3h_t handle, int row, int col, int len);
extern const char *sSnapAsciiBlock(l3h_t handle, int row, int col, int rows,
    int cols);
extern const char *sSnapEbcdicAll(l3h_t handle);
extern const char *sSnapEbcdicFromCursor(l3h_t handle, int len);
extern const char *sSnapEbcdicLine(l3h_t handle, int row, int col, int len);
extern const char *sSnapEbcdicBlock(l3h_t handle, int row, int col, int rows,
    int cols);
extern const char *sStatus(l3h_t handle);
extern int         aString(l3h_t handle, const char *);
extern int         aSysReq(l3h_t handle);
extern int         sTab(l3h_t handle);
extern int         sToggleInsert(l3h_t handle);
extern int         sToggleReverse(l3h_t handle);
extern int         aTransfer(l3h_t handle, int argc, const char *argv[]);
extern int         sUp(l3h_t handle);
extern int         aWaitInputField(l3h_t handle, int tmo);
extern int         aWaitOutput(l3h_t handle, int tmo);
extern int         aWait3270Mode(l3h_t handle, int tmo);
extern int         aWaitNVTMode(l3h_t handle, int tmo);
extern int         aWaitDisconnect(l3h_t handle, int tmo);

#endif /*]*/
