/* Test program for lib3270. */

#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <string.h>
#include "lib3270.h"

int use_select = 0;

static void doit(unsigned long flags, const char *hostname);
static void poll_it(l3h_t, const char *);

int
main(int argc, char *argv[])
{
	char *programname = argv[0];
	unsigned long flags = 0L;

	if (argc > 1 && !strcmp(argv[1], "-n")) {
		/* Use non-blocking mode in the library. */
		flags |= NON_BLOCKING;
		argc--;
		argv++;
	}
	if (argc > 1 && !strcmp(argv[1], "-s")) {
		/*
		 * Use explicit select() calls for non-blocking command
		 * completion.
		 */
		use_select++;
		argc--;
		argv++;
	}
	if (argc < 2) {
		fprintf(stderr, "usage: %s [-n] [-s] hostname\n", programname);
		exit(1);
	}

	/* Do it once. */
	doit(flags, argv[1]);

	/* Do it again. */
	doit(flags, argv[1]);
	return 0;
}

static void
doit(unsigned long flags, const char *hostname)
{
	static const char *no_opts[] = { NULL };
	l3h_t h;
	int rv;
	const char *r;

	/* Start up the library. */
	if (open_3270(flags, no_opts, &h) < 0) {
		fprintf(stderr, "open failed: %s\n", result_3270(h, NULL));
		exit(1);
	}

	/* Connect to the host. */
	if ((rv = aConnect(h, hostname)) < 0) {
		if (rv != L_BLOCKED) {
			fprintf(stderr, "Connect failed: %s\n",
			    result_3270(h, NULL));
			exit(1);
		}
		poll_it(h, "aConnect");
	}

	/* Wait for an input field. */
	if ((rv = aWaitInputField(h, 0)) < 0) {
		if (rv != L_BLOCKED) {
			fprintf(stderr, "WaitInputField failed: %s\n",
			    result_3270(h, NULL));
			exit(1);
		}
		poll_it(h, "aWaitInputField");
	}

	/* Get the contents of the screen. */
	r = sAsciiAll(h);
	if (r == NULL) {
		fprintf(stderr, "Ascii failed: %s\n", result_3270(h, NULL));
		exit(1);
	}
	printf("%s\n", r);

	/* Move the cursor up. */
	if (sUp(h) < 0) {
		fprintf(stderr, "Up failed: %s\n", result_3270(h, NULL));
		exit(1);
	}

	/* Done. */
	if (close_3270(h) < 0) {
		fprintf(stderr, "close failed: %s\n", result_3270(h, NULL));
		exit(1);
	}
}

/* Wait for lib3270 to complete a pending command. */
static void
poll_it(l3h_t h, const char *s)
{
	int p;

	if (use_select) {
		/* Use select() to do our own I/O multiplexing. */
		do {
			int nfds;
			fd_set r, w, x;
			struct timeval timebuf;
			struct timeval *timeout;
			int ns;

			nfds = 0;
			FD_ZERO(&r);
			FD_ZERO(&w);
			FD_ZERO(&x);
			timeout = NULL;
			/*
			 * If this were a real application, we would set up
			 * file descriptors and possibly a timeout.  Then
			 * selectvals_3270() would add its own.
			 */

			if (selectvals_3270(h, &nfds, &r, &w, &x, &timeout,
					    &timebuf) <= 0) {
				fprintf(stderr, "expected an event\n");
				exit(1);
			}
			ns = select(nfds, &r, &w, &x, timeout);
			if (ns < 0) {
				perror("select");
				exit(1);
			}
			printf("(got %d event%s)\n", ns, (ns==1)? "": "s");

			/*
			 * If this were a real application, we would examine
			 * our other file descriptors, and process any of our
			 * own timeouts that had occurred.
			 *
			 * Here however, we just call poll_3270(), which
			 * processes any pending I/Os and timeouts that
			 * lib3270 is interested in.
			 */
		} while ((p = poll_3270(h)) == L_BLOCKED);
	} else {
		/* Block in the library directly. */
		p = wait_3270(h);
	}
	if (p != L_OK) {
		fprintf(stderr, "'%s' failed: %s\n", s, result_3270(h, NULL));
		exit(1);
	}
}
